import os
import sys
import pytest
from pprint import pprint

# Add relevant ranger module to PATH... there surely is a better way to do this...
sys.path.insert(0, os.path.join(os.path.dirname(__file__), ".."))

from pymycobot import MyCobotCommandGenerator


@pytest.fixture(scope="module")
def mcg():
    print("")
    DEBUG = False
    mg = MyCobotCommandGenerator(debug=DEBUG)
    print("")
    return mg

def test_flatten1(mcg):
    assert mcg._flatten([1,2,3,4,5,6]) == [1,2,3,4,5,6]

def test_flatten2(mcg):
    assert mcg._flatten([1,2,3,4,[],5]) == [1,2,3,4,5]

def test_flatten3(mcg):
    assert mcg._flatten([1,2,3,4,[5,6],7]) == [1,2,3,4,5,6,7]

def test_encode_int16(mcg):
    assert mcg._encode_int16([0, 1]) == sum([mcg._encode_int16(0), mcg._encode_int16(1)], [])

def test_process_data_command1(mcg):
    assert mcg._process_data_command(()) == []

def test_process_data_command2(mcg):
    assert mcg._process_data_command((1, 2, 3)) == [1, 2, 3]

def test_process_data_command3(mcg):
    assert mcg._process_data_command((1, 2, [1, 2])) == mcg._flatten([1, 2, mcg._encode_int16([1, 2])])

def test_version(mcg):
    assert mcg.version() == ([0xFE, 0xFE, 2, 0x00, 0xFA], True)

def test_power_on(mcg):
    assert mcg.power_on() == ([0xFE, 0xFE, 2, 0x10, 0xFA], False)

def test_power_off(mcg):
    assert mcg.power_off() == ([0xFE, 0xFE, 2, 0x11, 0xFA], False)

def test_release_all_servos(mcg):
    assert mcg.release_all_servos() == ([0xFE, 0xFE, 2, 0x13, 0xFA], False)

def test_is_controller_connected(mcg):
    assert mcg.is_controller_connected() == ([0xFE, 0xFE, 2, 0x14, 0xFA], True)

def test_get_angles(mcg):
    assert mcg.get_angles() == ([0xFE, 0xFE, 2, 0x20, 0xFA], True)

def test_get_coords(mcg):
    assert mcg.get_coords() == ([0xFE, 0xFE, 2, 0x23, 0xFA], True)

def test_is_in_position(mcg):
    assert mcg.is_in_position([0, 0, 0, 0, 0, 0], 0) == (mcg._flatten([0xFE, 0xFE, 15, 0x2A, mcg._flatten([mcg._encode_int16([0, 0, 0, 0, 0, 0]), 0]), 0xFA]), True)

def test_pause(mcg):
    assert mcg.pause() == ([0xFE, 0xFE, 2, 0x26, 0xFA], False)

def test_resume(mcg):
    assert mcg.resume() == ([0xFE, 0xFE, 2, 0x28, 0xFA], False)

def test_stop(mcg):
    assert mcg.stop() == ([0xFE, 0xFE, 2, 0x29, 0xFA], False)

def test_set_pin_mode(mcg):
    assert mcg.set_pin_mode(1, 0) == ([0xFE, 0xFE, 4, 0x60, 1, 0, 0xFA], False)

def test_set_digital_output(mcg):
    assert mcg.set_digital_output(0, 1) == ([0xFE, 0xFE, 4, 0x61, 0, 1, 0xFA], False)



# coding=utf-8

from __future__ import division
import time
import struct
import sys


class ProtocolCode(object):
    # BASIC
    HEADER = 0xFE
    FOOTER = 0xFA

    # System status
    ROBOT_VERSION = 0x01
    SOFTWARE_VERSION = 0x02
    GET_ROBOT_ID = 0x03
    SET_ROBOT_ID = 0x04
    
    GET_ERROR_INFO = 0x07
    CLEAR_ERROR_INFO = 0x08
    GET_ATOM_VERSION = 0x09

    # Overall status
    POWER_ON = 0x10
    POWER_OFF = 0x11
    IS_POWER_ON = 0x12
    RELEASE_ALL_SERVOS = 0x13
    IS_CONTROLLER_CONNECTED = 0x14
    READ_NEXT_ERROR = 0x15
    SET_FRESH_MODE = 0x16
    GET_FRESH_MODE = 0x17
    SET_FREE_MODE = 0x1A
    IS_FREE_MODE = 0x1B
    COBOTX_GET_ANGLE = 0x1C
    COBOTX_IS_GO_ZERO = 0x1D
    FOCUS_ALL_SERVOS = 0x18
    GO_ZERO = 0x19

    # MDI MODE AND OPERATION
    GET_ANGLES = 0x20
    SEND_ANGLE = 0x21
    SEND_ANGLES = 0x22
    GET_COORDS = 0x23
    SEND_COORD = 0x24
    SEND_COORDS = 0x25
    PAUSE = 0x26
    IS_PAUSED = 0x27
    RESUME = 0x28
    STOP = 0x29
    IS_IN_POSITION = 0x2A
    IS_MOVING = 0x2B
    GET_ANGLE = 0x2C
    GET_COORD = 0x2D
    SEND_ANGLES_AUTO = 0x2E
    GET_SOLUTION_ANGLES = 0x2E
    SET_SOLUTION_ANGLES = 0x2F

    # JOG MODE AND OPERATION
    JOG_ANGLE = 0x30
    JOG_ABSOLUTE = 0x31
    JOG_COORD = 0x32
    JOG_INCREMENT = 0x33
    JOG_STOP = 0x34
    
    COBOTX_GET_SOLUTION_ANGLES = 0x35
    COBOTX_SET_SOLUTION_ANGLES = 0x36
    
    SET_ENCODER = 0x3A
    GET_ENCODER = 0x3B
    SET_ENCODERS = 0x3C
    GET_ENCODERS = 0x3D
    SET_ENCODERS_DRAG = 0x3E

    # RUNNING STATUS AND SETTINGS
    GET_SPEED = 0x40
    SET_SPEED = 0x41
    GET_FEED_OVERRIDE = 0x42
    GET_ACCELERATION = 0x44
    SET_ACCELERATION = 0x45
    GET_JOINT_MIN_ANGLE = 0x4A
    GET_JOINT_MAX_ANGLE = 0x4B
    SET_JOINT_MIN = 0x4C
    SET_JOINT_MAX = 0x4D

    # SERVO CONTROL
    IS_SERVO_ENABLE = 0x50
    IS_ALL_SERVO_ENABLE = 0x51
    SET_SERVO_DATA = 0x52
    GET_SERVO_DATA = 0x53
    SET_SERVO_CALIBRATION = 0x54
    JOINT_BRAKE = 0x55
    RELEASE_SERVO = 0x56
    FOCUS_SERVO = 0x57

    # ATOM IO
    SET_PIN_MODE = 0x60
    SET_DIGITAL_OUTPUT = 0x61
    GET_DIGITAL_INPUT = 0x62
    SET_PWM_MODE = 0x63
    SET_PWM_OUTPUT = 0x64
    GET_GRIPPER_VALUE = 0x65
    SET_GRIPPER_STATE = 0x66
    SET_GRIPPER_VALUE = 0x67
    SET_GRIPPER_CALIBRATION = 0x68
    IS_GRIPPER_MOVING = 0x69
    SET_COLOR = 0x6A
    SET_COLOR_MYARM = 0x70
    SET_ELETRIC_GRIPPER = 0x6B
    INIT_ELETRIC_GRIPPER = 0x6C
    SET_GRIPPER_MODE = 0x6D
    GET_GRIPPER_MODE = 0x6E

    GET_ACCEI_DATA = 0x73

    # Basic
    SET_BASIC_OUTPUT = 0xA0
    GET_BASIC_INPUT = 0xA1
    GET_BASE_INPUT = 0xA2
    SET_BASE_PWM = 0xA5

    # Linux GPIO, mode: GPIO.BCM
    SET_GPIO_MODE = 0xAA
    SET_GPIO_UP = 0xAB
    SET_GPIO_OUTPUT = 0xAC
    GET_GPIO_IN = 0xAD

    # set WIFI
    SET_SSID_PWD = 0xB0
    GET_SSID_PWD = 0xB1
    SET_SERVER_PORT = 0xB2

    # Get the measured distance
    GET_TOF_DISTANCE = 0xC0
    GET_BASIC_VERSION = 0xC1
    SET_COMMUNICATE_MODE = 0xC2
    GET_COMMUNICATE_MODE = 0xC3

    # Coordinate transformation
    SET_TOOL_REFERENCE = 0x81
    GET_TOOL_REFERENCE = 0x82
    SET_WORLD_REFERENCE = 0x83
    GET_WORLD_REFERENCE = 0x84
    SET_REFERENCE_FRAME = 0x85
    GET_REFERENCE_FRAME = 0x86
    SET_MOVEMENT_TYPE = 0x87
    GET_MOVEMENT_TYPE = 0x88
    SET_END_TYPE = 0x89
    GET_END_TYPE = 0x8A
    WRITE_MOVE_C = 0x8C

    # Impact checking
    SET_JOINT_CURRENT = 0x90
    GET_JOINT_CURRENT = 0x91
    SET_CURRENT_STATE = 0x92

    # planning speed
    GET_PLAN_SPEED = 0xD0
    GET_PLAN_ACCELERATION = 0xD1
    SET_PLAN_SPEED = 0xD2
    SET_PLAN_ACCELERATION = 0xD3
    move_round = 0xD4
    GET_ANGLES_COORDS = 0xD5

    # Motor status read
    GET_SERVO_SPEED = 0xE1
    GET_SERVO_CURRENTS = 0xE2
    GET_SERVO_VOLTAGES = 0xE3
    GET_SERVO_STATUS = 0xE4
    GET_SERVO_TEMPS = 0xE5
    GET_SERVO_LASTPDI = 0xE6
    SERVO_RESTORE = 0xE7

    GET_BASE_COORDS = 0xF0
    BASE_TO_SINGLE_COORDS = 0xF1
    COLLISION = 0xF2
    GET_BASE_COORD = 0xF3
    GET_ALL_BASE_COORDS = 0xF4
    WRITE_BASE_COORD = 0xF5
    WRITE_BASE_COORDS = 0xF6
    JOG_INC_COORD = 0xF7
    COLLISION_SWITCH = 0xF8
    IS_COLLISION_ON = 0xF9

    # IIC
    # SET_IIC_STATE = 0xA4
    # GET_IIS_BYTE = 0xA5
    # SET_IIC_BYTE = 0xA6

    # ultraArm
    END = "\r"
    COORDS_SET = "G0"
    SLEEP_TIME = "G4"
    BACK_ZERO = "G28"
    ABS_CARTESIAN = "G90"
    REL_CARTESIAN = "G91"
    SET_JOINT = "G92"
    SET_PWM = "M21"
    GPIO_CLOSE = "M23"
    GPIO_ON = "M22"
    GRIPPER_ZERO = "M24"
    GIRPPER_OPEN = "M25"
    GIRPPER_RELEASE = "M26"
    FAN_ON = "M27"
    FAN_CLOSE = "M28"
    SET_ANGLE = "M10"
    SET_ANGLES = "M11"
    GET_CURRENT_ANGLE_INFO = "M12"
    SET_JOG_ANGLE = "M13"
    JOG_COORD_ = "M14"
    SET_JOG_STOP = "M15"
    RELEASE_SERVOS = "M17"
    LOCK_SERVOS = "M18"
    GET_CURRENT_COORD_INFO = "M114"
    GET_BACK_ZERO_STATUS = "M119"
    IS_MOVING_END = "M9"


class DataProcessor(object):
    # Functional approach
    def _encode_int8(self, data):
        return struct.pack("b", data)

    def _encode_int16(self, data):
        if isinstance(data, int):
            return [
                ord(i) if isinstance(i, str) else i
                for i in list(struct.pack(">h", data))
            ]
        else:
            return sum([list(struct.pack(">h", elem)) for elem in data], [])

    def _decode_int8(self, data):
        return struct.unpack("B", data)[0]

    def _decode_int16(self, data):
        return struct.unpack(">h", data)[0]

    def _angle2int(self, angle):
        return int(angle * 100)

    def _coord2int(self, coord):
        return int(coord * 10)

    def _int2angle(self, _int):
        return round(_int / 100.0, 3)

    def _int2coord(self, _int):
        return round(_int / 10.0, 2)

    # def encode_int16(self, data):
    #     encoded_data = []

    #     if isinstance(data, int):
    #         encoded_data.extend(struct.pack(">h", data))
    #     else:
    #         for elem in data:
    #             encoded_data.extend(struct.pack(">h", elem))

    #     return [byte for byte in encoded_data]

    def _flatten(self, lst):
        flat_list = []
        for item in lst:
            if not isinstance(item, list):
                flat_list.append(item)
            else:
                flat_list.extend(self._flatten(item))
        return flat_list

    def _process_data_command(self, genre, _class, args):
        if not args:
            return []

        processed_args = []
        for index in range(len(args)):
            if isinstance(args[index], list):
                if genre == ProtocolCode.SET_ENCODERS_DRAG and index == 0 and _class == "CobotX":
                    for data in args[index]:
                        byte_value = data.to_bytes(4, byteorder='big', signed=True)
                        res = []
                        for i in range(len(byte_value)):
                           res.append(byte_value[i])
                        processed_args.extend(res)
                else:
                    processed_args.extend(self._encode_int16(args[index]))
            else:
                if isinstance(args[index], str):
                    processed_args.append(ord(args[index]))
                else:
                    if genre == ProtocolCode.SET_SERVO_DATA and _class == "CobotX" and index == 2:
                        byte_value = args[index].to_bytes(2, byteorder='big', signed=True)
                        res = []
                        for i in range(len(byte_value)):
                           res.append(byte_value[i])
                        processed_args.extend(res)
                    else:
                        processed_args.append(args[index])

        return processed_args

    def _is_frame_header(self, data, pos1, pos2):
        return data[pos1] == ProtocolCode.HEADER and data[pos2] == ProtocolCode.HEADER

    def _process_received(self, data, genre, arm=6):
        if not data:
            return []
        if genre == 177:
            data = str(data)[2:-1].split(": ")
            return data[1][0:-9], data[-1]
        if not data:
            return []
        data = bytearray(data)
        data_len = len(data)
        # Get valid header: 0xfe0xfe
        header_i, header_j = 0, 1
        while header_j < data_len - 4:
            if self._is_frame_header(data, header_i, header_j):
                if arm in [6, 7]:
                    cmd_id = data[header_i + 3]
                elif arm == 12:
                    cmd_id = data[header_i + 4]
                # compare send header and received header
                if cmd_id == genre:
                    break
            header_i += 1
            header_j += 1
        else:
            return []
        if arm in [6, 7]:
            data_len = data[header_i + 2] - 2
        elif arm == 12:
            data_len = data[header_i + 3] - 2
        unique_data = [ProtocolCode.GET_BASIC_INPUT, ProtocolCode.GET_DIGITAL_INPUT]

        if cmd_id in unique_data:
            if arm == 12:
                data_pos = header_i + 6
            else:
                data_pos = header_i + 5
            data_len -= 1
        else:
            if arm in [6, 7]:
                data_pos = header_i + 4
            elif arm == 12:
                data_pos = header_i + 5

        valid_data = data[data_pos : data_pos + data_len]

        # process valid data
        res = []
        if genre in [ProtocolCode.GET_SERVO_VOLTAGES, ProtocolCode.GET_SERVO_STATUS, ProtocolCode.GET_SERVO_TEMPS, ProtocolCode.GO_ZERO]:
            for i in valid_data:
                res.append(i)
            return res            
        if data_len in [6, 8, 12, 14, 24, 26, 60]:
            for header_i in range(0, len(valid_data), 2):
                one = valid_data[header_i : header_i + 2]
                res.append(self._decode_int16(one))
        elif data_len == 2:
            if genre in [
                ProtocolCode.GET_PLAN_SPEED,
                ProtocolCode.GET_PLAN_ACCELERATION,
            ]:
                return [
                    self._decode_int8(valid_data[0:1]),
                    self._decode_int8(valid_data[1:]),
                ]
            elif genre in [ProtocolCode.IS_SERVO_ENABLE]:
                return [self._decode_int8(valid_data[1:2])]
            elif genre in [ProtocolCode.GET_ERROR_INFO]:
                return [self._decode_int8(valid_data[1:])]
            res.append(self._decode_int16(valid_data))
        elif data_len == 3:
            res.append(self._decode_int16(valid_data[1:]))
        elif data_len == 4:
            for i in range(1,4):
                res.append(valid_data[i])
        elif data_len == 7:
            error_list = [i for i in valid_data]
            for i in error_list:
                if i in range(16,23):
                    res.append(1)
                elif i in range(23,29):
                    res.append(2)
                elif i in range(32,112):
                    res.append(3)
                else:
                    res.append(i)
        else:
            if genre in [
                ProtocolCode.GET_SERVO_VOLTAGES,
                ProtocolCode.GET_SERVO_STATUS,
                ProtocolCode.GET_SERVO_TEMPS,
            ]:
                for i in range(data_len):
                    data1 = self._decode_int8(valid_data[i : i + 1])
                    res.append(0xFF & data1 if data1 < 0 else data1)
                return res
            res.append(self._decode_int8(valid_data))
        if genre == ProtocolCode.GET_ACCEI_DATA:
            for i in range(len(res)):
                res[i] /= 1000
        return res

    def _process_single(self, data):
        return data[0] if data else -1

def check_python_version():
    if sys.version_info.major == 2:
        return 2
    elif sys.version_info.major == 3:
        return 3
    else:
        return -1


def write(self, command, method=None):
    if len(command) > 3 and command[3] == 176 and len(command) > 5:
        command = "'" + command[4] + "'" + "(" + command[5] + ")"
        command = command.encode()
    if method == "socket":
        self.log.debug("_write: {}".format([hex(i) for i in command]))
        py_version = check_python_version()
        if py_version == 2:
            self.sock.sendall("".join([chr(b) for b in command]))
        else:
            self.sock.sendall(bytes(command))
    else:
        self._serial_port.reset_input_buffer()
        self.log.debug("_write: {}".format([hex(i) for i in command]))
        self._serial_port.write(command)
        self._serial_port.flush()


def read(self, genre, method=None, command=None):
    datas = b""
    data_len = -1
    k = 0
    pre = 0
    t = time.time()
    wait_time = 0.1    
    if genre == ProtocolCode.GO_ZERO:
        wait_time = 120
    if method is not None:
        if genre == 177:
            while True:
                data = self.sock.recv(1024)
                if b"password" in data:
                    break
        elif genre == 192:
            while True:
                data += self.sock.recv(1024)
                if len(data) == 6:
                    break
        else:
            try:
                self.sock.settimeout(0.5)
                data = self.sock.recv(1024)
                if isinstance(data, str):
                    datas = bytearray()
                    for i in data:   
                        datas += hex(ord(i))
            except:
                data = b""
        if check_python_version() == 2:
            self.log.debug("_read: {}".format([hex(ord(d)) for d in data]))
        else:
            self.log.debug("_read: {}".format([hex(d) for d in data]))
        return data
    else:
        if genre == ProtocolCode.GET_SSID_PWD:
            time.sleep(0.1)
            if self._serial_port.inWaiting() > 0:
                datas = self._serial_port.read(self._serial_port.inWaiting())
            return datas
        elif genre == ProtocolCode.GET_ACCEI_DATA:
            wait_time = 1
        while True and time.time() - t < wait_time:
            data = self._serial_port.read()
            k += 1
            if data_len == 1 and data == b"\xfa":
                datas += data
                if [i for i in datas] == command:
                    datas = b''
                    data_len = -1
                    k = 0
                    pre = 0
                    continue
                break
            elif len(datas) == 2:
                data_len = struct.unpack("b", data)[0]
                datas += data
            elif len(datas) > 2 and data_len > 0:
                datas += data
                data_len -= 1
            elif data == b"\xfe":
                if datas == b"":
                    datas += data
                    pre = k
                else:
                    if k - 1 == pre:
                        datas += data
                    else:
                        datas = b"\xfe"
                        pre = k
        else:
            datas = b''
        self.log.debug("_read: {}".format([hex(ord(data)) for data in datas]))
        return datas



# coding=utf-8

from __future__ import division
import time
import math
import logging
import struct

from pymycobot.log import setup_logging
from pymycobot.Interface import MyBuddyCommandGenerator
from pymycobot.common import ProtocolCode, write, read
from pymycobot.error import calibration_parameters


class MyBuddy(MyBuddyCommandGenerator):
    """MyCobot Python API Serial communication class.

    Supported methods:

        # Overall status
            Look at parent class: `CommandGenerator`.

        # MDI mode and operation
            get_radians()
            send_radians()
            sync_send_angles() *
            sync_send_coords() *
            Other look at parent class: `CommandGenerator`.

        # JOG mode and operation
            Look at parent class: `CommandGenerator`.

        # Running status and Settings
            Look at parent class: `CommandGenerator`.

        # Servo control
            Look at parent class: `CommandGenerator`.

        # Atom IO
            Look at parent class: `CommandGenerator`.

        # Basic
            Look at parent class: `CommandGenerator`.

        # Other
            wait() *
    """

    def __init__(self, port, baudrate="115200", timeout=0.1, debug=False):
        """
        Args:
            port     : port string
            baudrate : baud rate string, default '115200'
            timeout  : default 0.1
            debug    : whether show debug info
        """
        super(MyBuddy, self).__init__(debug)
        self.calibration_parameters = calibration_parameters
        import serial
        self._serial_port = serial.Serial()
        self._serial_port.port = port
        self._serial_port.baudrate = baudrate
        self._serial_port.timeout = timeout
        self._serial_port.rts = False
        self._serial_port.open()

    _write = write
    
    def _read(self):
        datas = b''
        data_len = -1
        real_data = b''
        check_digit = 0
        k = 0
        pre = 0
        t = time.time()
        while True and time.time() - t < 0.1:
            try:
                data = self._serial_port.read()
                k+=1
                if data_len == 1:
                    datas += data
                    if struct.pack("B", check_digit & 0xff) == data:
                        break
                elif len(datas) == 3:
                    data_len = struct.unpack("b",data)[0]
                    datas += data
                elif len(datas)>=2:
                    datas += data
                    if len(datas) == 5: 
                        check_digit += self._decode_int8(data)
                    elif len(datas)>5:
                        real_data += data
                    if real_data != b'':
                        check_digit += self._decode_int8(real_data)
                        real_data = b''
                    data_len -= 1
                elif data == b'\xfe':
                    if datas == b'':
                        datas += data
                        pre = k
                    else:
                        if k-1 == pre:
                            datas += data
                        else:
                            datas = b'\xfe'
                            pre = k  
            except:
                self.log.debug("_read: {}".format(datas))
                
                datas = None
                break
        else:
            self.log.debug("_read: {}".format(datas))

            datas = None
        self.log.debug("_read: {}".format(datas))
        return datas
    
    def _mesg(self, genre, *args, **kwargs):
        """

        Args:
            genre: command type (Command)
            *args: other data.
                   It is converted to octal by default.
                   If the data needs to be encapsulated into hexadecimal,
                   the array is used to include them. (Data cannot be nested)
            **kwargs: support `has_reply`
                has_reply: Whether there is a return value to accept.
        """
        real_command, has_reply = super(
            MyBuddy, self)._mesg(genre, *args, **kwargs)
        self._write(self._flatten(real_command))

        if has_reply:
            data = self._read()
            res = self._process_received(data, genre, arm=12)
            if genre in [
                ProtocolCode.ROBOT_VERSION,
                ProtocolCode.SOFTWARE_VERSION,
                ProtocolCode.GET_ROBOT_ID,
                ProtocolCode.IS_POWER_ON,
                ProtocolCode.IS_CONTROLLER_CONNECTED,
                ProtocolCode.IS_PAUSED,  # TODO have bug: return b''
                ProtocolCode.IS_IN_POSITION,
                ProtocolCode.IS_MOVING,
                ProtocolCode.IS_SERVO_ENABLE,
                ProtocolCode.IS_ALL_SERVO_ENABLE,
                ProtocolCode.GET_SERVO_DATA,
                ProtocolCode.GET_DIGITAL_INPUT,
                ProtocolCode.GET_GRIPPER_VALUE,
                ProtocolCode.IS_GRIPPER_MOVING,
                ProtocolCode.GET_SPEED,
                ProtocolCode.GET_ENCODER,
                ProtocolCode.GET_BASIC_INPUT,
                ProtocolCode.GET_TOF_DISTANCE,
                ProtocolCode.GET_END_TYPE,
                ProtocolCode.GET_MOVEMENT_TYPE,
                ProtocolCode.GET_REFERENCE_FRAME,
                ProtocolCode.GET_JOINT_MIN_ANGLE,
                ProtocolCode.GET_JOINT_MAX_ANGLE,
                ProtocolCode.GET_FRESH_MODE,
                # ProtocolCode.GET_SERVO_CURRENTS
            ]:
                return self._process_single(res)
            elif genre in [ProtocolCode.GET_ANGLES]:
                return [self._int2angle(angle) for angle in res]
            elif genre in [ProtocolCode.GET_ANGLE]:
                return self._int2angle(res[0]) if res else None
            elif genre in [ProtocolCode.GET_COORD]:
                if real_command[5] < 4:
                    if real_command[2] == 3:
                        return self._int2angle(res[0]) if res else None
                    return self._int2coord(res[0]) if res else None
                else:
                    return self._int2angle(res[0]) if res else None
            elif genre in [ProtocolCode.GET_ALL_BASE_COORDS, ProtocolCode.GET_COORDS, ProtocolCode.GET_TOOL_REFERENCE, ProtocolCode.GET_WORLD_REFERENCE, ProtocolCode.GET_BASE_COORDS, ProtocolCode.GET_BASE_COORD, ProtocolCode.BASE_TO_SINGLE_COORDS]:
                if res:
                    r = [] 
                    for idx in range(3):
                        r.append(self._int2coord(res[idx]))
                    for idx in range(3, 6):
                        r.append(self._int2angle(res[idx]))
                    if len(res) == 12:
                        r1 = []
                        for idx in range(6, 9):
                            r1.append(self._int2coord(res[idx]))
                        for idx in range(9, 12):
                            r1.append(self._int2angle(res[idx]))
                        return [r, r1]
                    return r
                else:
                    return res
            elif genre in [ProtocolCode.GET_SERVO_VOLTAGES, ProtocolCode.COLLISION]:
                return [self._int2coord(angle) for angle in res]
            else:
                return res
        return None

    def get_radians(self, id):
        """Get the radians of all joints

        Args: 
            id: 1/2 (L/R)
            
        Return:
            list: A list of float radians [radian1, ...]
        """
        angles = self._mesg(ProtocolCode.GET_ANGLES, id, has_reply=True)
        return [round(angle * (math.pi / 180), 3) for angle in angles]

    def send_radians(self, id, radians, speed):
        """Send the radians of all joints to robot arm

        Args:
            id: 1/2 (L/R).
            radians: a list of radian values( List[float]), length 6
            speed: (int )0 ~ 100
        """
        # calibration_parameters(len6=radians, speed=speed)
        degrees = [self._angle2int(radian * (180 / math.pi))
                   for radian in radians]
        return self._mesg(ProtocolCode.SEND_ANGLES, id, degrees, speed)

    # Basic for raspberry pi.
    def set_gpio_init_mode(self, mode = 1):
        """import RPiGPIO,Init GPIO mode
        
            Args:
                mode(int):0/1
                0 = BCM
                1 = BOAND
        """
        import RPi.GPIO as GPIO  # type: ignore
        self.gpio = GPIO
        if mode == 0:
            self.gpio.setmode(GPIO.BCM)
        else:
            self.gpio.setmode(GPIO.BOAND)

    def set_gpio_setup(self, pin_no, mode):
        """Init GPIO module, and set BCM mode. 

        Args:
            pin_no: (int)pin number 1-16.
        
                PIN_NO = GPIO.BCM:  

                |1 = G7  |  2 = G8  |  3 = G25 | 4 = G24  | 5 = G23  | 6 = G18   |

                |7 = G11 |  8 = G9  |  9 = G10 | 10 =G22  | 11 =G27  | 12 = G17  |

                |Grove0:   |  SCL0 = 13 = G3    |     SDA0 = 14 = G2    |

                |Grove1:   |  SCL1 = 15 = G6    |     SDA1 = 16 = G5    |

            mode: 
                0 - input
                    define: pull_up_down = DOWN
                1 - output
                     define: initial = HIGH
        """
        pin_no = self.base_io_to_gpio(pin_no)       
        if mode == 1:
            self.gpio.setup(pin_no, self.gpio.OUT)
        else:
            self.gpio.setup(pin_no, self.gpio.IN)
            
    def set_gpio_output(self, pin_no, v):
        """Set GPIO output value.

        Args:
            pin_no: (int)pin number 1-16.
            v: (int) 0 / 1
        """
        pin = self.base_io_to_gpio(pin_no)
        self.gpio.output(pin, v)
        
    def get_gpio_input(self, pin_no):
        """Get GPIO input value.

        Args:
            pin_no: (int)pin number 1-16.
        """
        pin = self.base_io_to_gpio(pin_no)
        return self.gpio.input(pin)
        
    def set_gpio_pwm_start(self, pin_no, freq = 0.5, dc = 0.5):
        """Set GPIO PWM value.

        Args:
            pin_no: (int)pin number 1-16.
            freq: (float) 0.0 - 1000000.0
            dc: (float) 0.0 - 100.0
        """
        pin = self.base_io_to_gpio(pin_no)
        self.pwm = self.gpio.PWM(pin, freq)
        self.pwm.start(dc)

    def set_gpio_pwm_change_freq(self, freq):
        """Reset GPIO PWM freq.

        Args:
            freq: (float) 0.0 - 1000000.0
        """
        self.pwm.ChangeFrequency(freq)

    def set_gpio_pwm_change_dc(self, dc):
        """Reset GPIO PWM DC.

        Args:
            dc: (float) 0.0 - 100.0
        """
        self.pwm.ChangeDutyCycle(dc)

    def set_gpio_pwm_stop(self):
        """Set GPIO PWM STOP OUTPUT.
        """
        self.pwm.stop()

    def set_iic_init(self, IIC_NO):
        """
            import SMBUS2

            Open IIC_NO port

            (For more use, please see
            
            pypilink: https://pypi.org/project/smbus2/,
             
            githublink: https://github.com/kplindegaard/smbus2)
            
        Args:
            IIC_NO(int) : 0/1 ,0 = iic0, 1=iic1
        """
        from smbus2 import SMBus
        # self.iic = SMBus(IIC_NO)
        return SMBus(IIC_NO)

    def base_io_to_gpio(self, pin):
        """BASE_io = GPIO.BCM:   
            1 = G7    |   7 = G11  
            2 = G8    |   8 = G9  
            3 = G25  |   9 = G10  
            4 = G24  |  10 =G22  
            5 = G23  |  11 =G27  
            6 = G18  |  12 = G17  
            GND       |  3V3  

            Grove0:
            SCL0 = 13 = G3  
            SDA0 = 14 = G2  
            5V  
            GND  

            Grove1:  
            SCL1 = 15 = G6  
            SDA1 = 16 = G5  
            5V  
            GND  
        """
        pin_dist = {1:7, 2:8, 3:25, 4:24, 5:23, 6:18, 
                    7:11, 8:9, 9:10, 10:22, 11:27, 12:17,
                    13:3, 14:2,'SCL0':3,"SDA0":2, 
                    15:6, 16:5, 'SCL0':6,"SDA0":5}
        if pin in pin_dist:
            _pin = pin
            pin = pin_dist.get(_pin)
        else:
            print('The IO definition exceeds the system support range, and the program exits automatically.')
            pin = None
        return pin

    def set_gpio_clearup(self,pin_no = None):
        """SET GPIO CLEANUP

        Args:
            pin_no :(int)pin number 1-16.
            if pin_no = None : cleanup all gpio
        """
        if pin_no:
            pin_no = self.base_io_to_gpio(pin_no)
        self.gpio.cleanup(pin_no)
   
    # Other
    def wait(self, t):
        time.sleep(t)
        return self

    def close(self):
        self._serial_port.close()   
            
    def open(self):
        self._serial_port.open()
        
    def is_open(self):
        return self._serial_port.is_open



# coding=utf-8

from __future__ import division
import time
import math
import logging

from pymycobot.log import setup_logging
from pymycobot.generate import CommandGenerator
from pymycobot.public import PublicCommandGenerator
from pymycobot.common import ProtocolCode, write, read
from pymycobot.error import calibration_parameters


class MyCobot(CommandGenerator, PublicCommandGenerator):
    """MyCobot Python API Serial communication class.

    Supported methods:

        # Overall status
            Look at parent class: `CommandGenerator`.

        # MDI mode and operation
            get_radians()
            send_radians()
            sync_send_angles() *
            sync_send_coords() *
            Other look at parent class: `CommandGenerator`.

        # JOG mode and operation
            Look at parent class: `CommandGenerator`.

        # Running status and Settings
            Look at parent class: `CommandGenerator`.

        # Servo control
            Look at parent class: `CommandGenerator`.

        # Atom IO
            Look at parent class: `CommandGenerator`.

        # Basic
            Look at parent class: `CommandGenerator`.

        # Other
            wait() *
    """

    def __init__(self, port, baudrate="115200", timeout=0.1, debug=False):
        """
        Args:
            port     : port string
            baudrate : baud rate string, default '115200'
            timeout  : default 0.1
            debug    : whether show debug info
        """
        super(MyCobot, self).__init__(debug)
        self.calibration_parameters = calibration_parameters
        import serial
        self._serial_port = serial.Serial()
        self._serial_port.port = port
        self._serial_port.baudrate = baudrate
        self._serial_port.timeout = timeout
        self._serial_port.rts = False
        self._serial_port.open()

    _write = write
    _read = read

    def _mesg(self, genre, *args, **kwargs):
        """

        Args:
            genre: command type (Command)
            *args: other data.
                   It is converted to octal by default.
                   If the data needs to be encapsulated into hexadecimal,
                   the array is used to include them. (Data cannot be nested)
            **kwargs: support `has_reply`
                has_reply: Whether there is a return value to accept.
        """
        real_command, has_reply = super(
            MyCobot, self)._mesg(genre, *args, **kwargs)
        self._write(self._flatten(real_command))

        if has_reply:
            data = self._read(genre)
            if genre == ProtocolCode.SET_SSID_PWD:
                return None
            res = self._process_received(data, genre)
            if genre in [
                ProtocolCode.ROBOT_VERSION,
                ProtocolCode.GET_ROBOT_ID,
                ProtocolCode.IS_POWER_ON,
                ProtocolCode.IS_CONTROLLER_CONNECTED,
                ProtocolCode.IS_PAUSED,  # TODO have bug: return b''
                ProtocolCode.IS_IN_POSITION,
                ProtocolCode.IS_MOVING,
                ProtocolCode.IS_SERVO_ENABLE,
                ProtocolCode.IS_ALL_SERVO_ENABLE,
                ProtocolCode.GET_SERVO_DATA,
                ProtocolCode.GET_DIGITAL_INPUT,
                ProtocolCode.GET_GRIPPER_VALUE,
                ProtocolCode.IS_GRIPPER_MOVING,
                ProtocolCode.GET_SPEED,
                ProtocolCode.GET_ENCODER,
                ProtocolCode.GET_BASIC_INPUT,
                ProtocolCode.GET_TOF_DISTANCE,
                ProtocolCode.GET_END_TYPE,
                ProtocolCode.GET_MOVEMENT_TYPE,
                ProtocolCode.GET_REFERENCE_FRAME,
                ProtocolCode.GET_FRESH_MODE,
                ProtocolCode.GET_GRIPPER_MODE,
                ProtocolCode.GET_ERROR_INFO
            ]:
                return self._process_single(res)
            elif genre in [ProtocolCode.GET_ANGLES]:
                return [self._int2angle(angle) for angle in res]
            elif genre in [ProtocolCode.GET_COORDS, ProtocolCode.GET_TOOL_REFERENCE, ProtocolCode.GET_WORLD_REFERENCE]:
                if res:
                    r = []
                    for idx in range(3):
                        r.append(self._int2coord(res[idx]))
                    for idx in range(3, 6):
                        r.append(self._int2angle(res[idx]))
                    return r
                else:
                    return res
            elif genre in [ProtocolCode.GET_SERVO_VOLTAGES]:
                return [self._int2coord(angle) for angle in res]
            elif genre in [ProtocolCode.GET_JOINT_MAX_ANGLE, ProtocolCode.GET_JOINT_MIN_ANGLE]:
                return self._int2coord(res[0])
            elif genre in [ProtocolCode.GET_BASIC_VERSION, ProtocolCode.SOFTWARE_VERSION, ProtocolCode.GET_ATOM_VERSION]:
                return self._int2coord(self._process_single(res))
            elif genre == ProtocolCode.GET_ANGLES_COORDS:
                r = []
                for index in range(len(res)):
                    if index < 6:
                        r.append(self._int2angle(res[index]))
                    elif index < 9:
                        r.append(self._int2coord(res[index]))
                    else:
                        r.append(self._int2angle(res[index]))
                return r
            else:
                return res
        return None

    def get_radians(self):
        """Get the radians of all joints

        Return:
            list: A list of float radians [radian1, ...]
        """
        angles = self._mesg(ProtocolCode.GET_ANGLES, has_reply=True)
        return [round(angle * (math.pi / 180), 3) for angle in angles]

    def send_radians(self, radians, speed):
        """Send the radians of all joints to robot arm

        Args:
            radians: a list of radian values( List[float]), length 6
            speed: (int )0 ~ 100
        """
        calibration_parameters(len6=radians, speed=speed)
        degrees = [self._angle2int(radian * (180 / math.pi))
                   for radian in radians]
        return self._mesg(ProtocolCode.SEND_ANGLES, degrees, speed)

    def sync_send_angles(self, degrees, speed, timeout=15):
        """Send the angle in synchronous state and return when the target point is reached
            
        Args:
            degrees: a list of degree values(List[float]), length 6.
            speed: (int) 0 ~ 100
            timeout: default 7s.
        """
        t = time.time()
        self.send_angles(degrees, speed)
        while time.time() - t < timeout:
            f = self.is_in_position(degrees, 0)
            if f == 1:
                break
            time.sleep(0.1)
        return self

    def sync_send_coords(self, coords, speed, mode=0, timeout=15):
        """Send the coord in synchronous state and return when the target point is reached
            
        Args:
            coords: a list of coord values(List[float])
            speed: (int) 0 ~ 100
            mode: (int): 0 - angular（default）, 1 - linear
            timeout: default 7s.
        """
        t = time.time()
        self.send_coords(coords, speed, mode)
        while time.time() - t < timeout:
            if self.is_in_position(coords, 1) == 1:
                break
            time.sleep(0.1)
        return self

    # Basic for raspberry pi.
    def gpio_init(self):
        """Init GPIO module, and set BCM mode."""
        import RPi.GPIO as GPIO  # type: ignore

        GPIO.setmode(GPIO.BCM)
        self.gpio = GPIO

    def gpio_output(self, pin, v):
        """Set GPIO output value.

        Args:
            pin: (int)pin number.
            v: (int) 0 / 1
        """
        self.gpio.setup(pin, self.gpio.OUT)
        self.gpio.output(pin, v)

    # Other
    def wait(self, t):
        time.sleep(t)
        return self
    
    def close(self):
        self._serial_port.close()
        
    def open(self):
        self._serial_port.open()
        
    def get_acceleration(self):
        """get acceleration"""
        return self._process_single(
            self._mesg(ProtocolCode.GET_ACCELERATION, has_reply=True)
        )

    def set_acceleration(self, acc):
        """Set speed for all moves
        
        Args:
            acc: int
        """
        return self._mesg(ProtocolCode.SET_ACCELERATION, acc)



import enum
import serial
import time
import struct
import logging
from pymycobot.log import setup_logging
from pymycobot.common import DataProcessor
from pymycobot.error import calibration_parameters

class ProtocolCode(enum.Enum):
    HEADER = 0xFE
    SET_LED = [0x01, 0x02]
    GET_FIRMWARE_VERSION = [0x01, 0x03]
    GET_MOTORS_CURRENT = [0x01, 0x04]
    GET_BATTERY_INFO = [0x01, 0x05]

class MyAgv(DataProcessor):
    def __init__(self, port="/dev/ttyAMA0", baudrate="115200", timeout=0.1, debug=False):
        self.debug = debug
        setup_logging(self.debug)
        self.log = logging.getLogger(__name__)
        self._serial_port = serial.Serial()
        self._serial_port.port = port
        self._serial_port.baudrate = baudrate
        self._serial_port.timeout = timeout
        self._serial_port.rts = False
        self._serial_port.open()
        
    def _write(self, command):
        self._serial_port.reset_input_buffer()
        self.log.debug("_write: {}".format([hex(i) for i in command]))
        self._serial_port.write(command)
        self._serial_port.flush()
        
    def _read(self, command):
        datas = b''
        k = 0
        pre = 0
        end = 5
        t = time.time()
        while time.time() - t < 0.2:
            data = self._serial_port.read()
            k += 1
            # print("data_len:",end)
            # print(data, len(datas), k, end)
            if len(datas) == 4:
                # print("------:",datas, command, k, end)
                if datas[-2] == 0x01 and datas[-1] == 0x05:
                    end = 7
                datas += data
            
                
            elif len(datas) == end:
                datas += data
                break
            elif len(datas) > 4:
                datas += data
            
            elif len(datas) >= 2:
                data_len = struct.unpack("b", data)[0]
                # print("``````:",datas, command, k, data_len)
                if data_len == command[k-1]:
                    datas += data
                else:
                    datas = b''
                    k = 0
                    pre = 0
            elif data == b"\xfe":
                if datas == b'':
                    datas += data
                    if k != 1:
                        k = 1
                    pre = k
                else:
                    if k - 1 == pre:
                        datas += data
                    else:
                        datas = b"\xfe"
                        k = 1
                        pre = 0
        else:
            datas = b''
        self.log.debug("_read: {}".format([hex(data) for data in datas]))
        return datas
            
        
        
    def _mesg(self, genre, *args, **kwargs):
        """

        Args:
            genre: command type (Command)
            *args: other data.
                   It is converted to octal by default.
                   If the data needs to be encapsulated into hexadecimal,
                   the array is used to include them. (Data cannot be nested)
            **kwargs: support `has_reply`
                has_reply: Whether there is a return value to accept.
        """
        has_reply = kwargs.get("has_reply", None)
        real_command = self._process_data_command(genre, self.__class__.__name__, args)
        command = [
            ProtocolCode.HEADER.value,
            ProtocolCode.HEADER.value,
        ]
        if isinstance(genre, list):
            for data in genre:
                command.append(data)
        else:
            command.append(genre)
        command.append(real_command)
        command = self._flatten(command)
        if genre == ProtocolCode.SET_LED.value:
            command.append(sum(command[2:]) & 0xff)
        elif genre == ProtocolCode.GET_FIRMWARE_VERSION.value:
            command.append(4)
        elif genre == ProtocolCode.GET_MOTORS_CURRENT.value:
            command.append(5)
        elif genre == ProtocolCode.GET_BATTERY_INFO.value:
            command.append(6)
        else:
            # del command[2]
            command.append(sum(command[2:]) & 0xff)
        self._write(command)
        if has_reply:
            data = self._read(command)
            if data:
                if genre == ProtocolCode.GET_FIRMWARE_VERSION.value:
                    return self._int2coord(data[4])
                elif genre == ProtocolCode.GET_MOTORS_CURRENT.value:
                    return self._decode_int16(data[4:6])
                elif ProtocolCode.GET_BATTERY_INFO.value:
                    byte_1 = bin(data[4])[2:]
                    while len(byte_1) != 6:
                        byte_1 = "0"+byte_1
                    return [byte_1, self._int2coord(data[5]), self._int2coord(data[6])]
            # print(res)
        return None
    
    def set_led(self, mode, R, G, B):
        """Set up LED lights

        Args:
            mode (int): 1 - Set LED light color. 2 - Set the LED light to blink
            R (int): 0 ~ 255
            G (int): 0 ~ 255
            B (int): 0 ~ 255
        """
        calibration_parameters(class_name = self.__class__.__name__, rgb=[R,G,B], led_mode=mode)
        return self._mesg(ProtocolCode.SET_LED.value, mode, R, G, B)
    
    def get_firmware_version(self):
        """Get firmware version number
        """
        return self._mesg(ProtocolCode.GET_FIRMWARE_VERSION.value, has_reply = True)
    
    def get_motors_current(self):
        """Get the total current of the motor
        """
        return self._mesg(ProtocolCode.GET_MOTORS_CURRENT.value, has_reply = True)
    
    def get_battery_info(self):
        """Read battery information
        
        Return:
            list : [battery_data, battery_1_voltage, battery_2_voltage].
                battery_data:
                    A string of length 6, represented from left to right: 
                    bit5, bit4, bit3, bit2, bit1, bit0.

                    bit5 : Battery 2 is inserted into the interface 1 means inserted, 0 is not inserted.
                    bit4 : Battery 1 is inserted into the interface, 1 means inserted, 0 is not inserted.
                    bit3 : The adapter is plugged into the interface 1 means plugged in, 0 not plugged in.
                    bit2 : The charging pile is inserted into the interface, 1 means plugged in, 0 is not plugged in.
                    bit1 : Battery 2 charging light 0 means off, 1 means on.
                    bit0 : Battery 1 charging light, 0 means off, 1 means on.
                battery_1_voltage : Battery 1 voltage in volts.
                battery_2_voltage : Battery 2 voltage in volts.
        """
        return self._mesg(ProtocolCode.GET_BATTERY_INFO.value, has_reply = True)
    
    def move_control(self, direction_1, direction_2, direction_3):
        """Control the car to rotate forward, backward, left, right and forward/counterclockwise

        Args:
            direction_1 (int): Control forward or backward: 0 ~ 127 is backward, 129 ~ 255 is forward, 128 is stop.
            direction_2 (int): control left and right movement: 0 ~ 127 is right, 129 ~ 255 is left, 128 is stop.
            direction_3 (int): control rotation: 0 ~ 127 is clockwise, 129 ~ 255 is counterclockwise, 128 is stop.
        """
        calibration_parameters(class_name = self.__class__.__name__, direction_1=direction_1, direction_2=direction_2,direction_3=direction_3)
        return self._mesg(direction_1, direction_2, direction_3)
    
    def go_ahead(self, go_speed):
        """Control the car to move forward

        Args:
            go_speed (int): 1 ~ 127 is forward
        """
        # go_speed (int): 129 ~ 255 is forward
        calibration_parameters(class_name = self.__class__.__name__, data=go_speed)
        self._mesg(128+go_speed, 128, 128)
        
    def retreat(self, back_speed):
        """Control the car back

        Args:
            back_speed (int): 1 ~ 127 is backward
        """
        calibration_parameters(class_name = self.__class__.__name__, data=back_speed)
        self._mesg(128-back_speed, 128, 128)
        
    def pan_left(self, pan_left_speed):
        """Control the car to pan to the left

        Args:
            pan_left_speed (int): 1 ~ 127
        """
        # pan_left_speed (int): 129 ~ 255
        calibration_parameters(class_name = self.__class__.__name__, data=pan_left_speed)
        self._mesg(128, 128+pan_left_speed, 128)
        
    def pan_right(self, pan_right_speed):
        """Control the car to pan to the right

        Args:
            pan_right_speed (int): 1 ~ 127
        """
        calibration_parameters(class_name = self.__class__.__name__, pan_right_speed=pan_right_speed)
        self._mesg(128, 128-pan_right_speed, 128)
        
    def clockwise_rotation(self, rotate_right_speed):
        """Control the car to rotate clockwise

        Args:
            clockwise_rotation_speed (int): 1 ~ 127
        """
        calibration_parameters(class_name = self.__class__.__name__, rotate_right_speed=rotate_right_speed)
        self._mesg(128, 128, 128-rotate_right_speed)
        
    def counterclockwise_rotation(self, rotate_left_speed):
        """Control the car to rotate counterclockwise

        Args:
            clockwise_rotation_speed (int): 1 ~ 127
        """
        calibration_parameters(class_name = self.__class__.__name__, rotate_left_speed=rotate_left_speed)
        self._mesg(128, 128, 128+rotate_left_speed)
        
    def stop(self):
        """stop motion
        """
        self._mesg(128, 128, 128)
    


# coding=utf-8

import time

from pymycobot.generate import CommandGenerator
from pymycobot.common import ProtocolCode, write, read
from pymycobot.error import calibration_parameters


class CobotX(CommandGenerator):
    _write = write
    _read = read
    def __init__(self, port, baudrate="115200", timeout=0.1, debug=False):
        """
        Args:
            port     : port string
            baudrate : baud rate string, default '115200'
            timeout  : default 0.1
            debug    : whether show debug info
        """
        super(CobotX, self).__init__(debug)
        self.calibration_parameters = calibration_parameters
        import serial

        self._serial_port = serial.Serial()
        self._serial_port.port = port
        self._serial_port.baudrate = baudrate
        self._serial_port.timeout = timeout
        self._serial_port.rts = False
        self._serial_port.open()
        

    def _mesg(self, genre, *args, **kwargs):
        """

        Args:
            genre: command type (Command)
            *args: other data.
                   It is converted to octal by default.
                   If the data needs to be encapsulated into hexadecimal,
                   the array is used to include them. (Data cannot be nested)
            **kwargs: support `has_reply`
                has_reply: Whether there is a return value to accept.
        """
        real_command, has_reply = super(CobotX, self)._mesg(genre, *args, **kwargs)
        self._write(self._flatten(real_command))

        if has_reply:
            data = self._read(genre)
            if genre == ProtocolCode.SET_SSID_PWD:
                return None
            res = self._process_received(data, genre)
            if genre in [
                ProtocolCode.ROBOT_VERSION,
                ProtocolCode.GET_ROBOT_ID,
                ProtocolCode.IS_POWER_ON,
                ProtocolCode.IS_CONTROLLER_CONNECTED,
                ProtocolCode.IS_PAUSED,  # TODO have bug: return b''
                ProtocolCode.IS_IN_POSITION,
                ProtocolCode.IS_MOVING,
                ProtocolCode.IS_SERVO_ENABLE,
                ProtocolCode.IS_ALL_SERVO_ENABLE,
                ProtocolCode.GET_SERVO_DATA,
                ProtocolCode.GET_DIGITAL_INPUT,
                ProtocolCode.GET_GRIPPER_VALUE,
                ProtocolCode.IS_GRIPPER_MOVING,
                ProtocolCode.GET_SPEED,
                ProtocolCode.GET_ENCODER,
                ProtocolCode.GET_BASIC_INPUT,
                ProtocolCode.GET_TOF_DISTANCE,
                ProtocolCode.GET_END_TYPE,
                ProtocolCode.GET_MOVEMENT_TYPE,
                ProtocolCode.GET_REFERENCE_FRAME,
                ProtocolCode.GET_FRESH_MODE,
                ProtocolCode.GET_GRIPPER_MODE,
                ProtocolCode.SET_SSID_PWD,
                ProtocolCode.COBOTX_IS_GO_ZERO,
            ]:
                return self._process_single(res)
            elif genre in [ProtocolCode.GET_ANGLES]:
                return [self._int2angle(angle) for angle in res]
            elif genre in [
                ProtocolCode.GET_COORDS,
                ProtocolCode.GET_TOOL_REFERENCE,
                ProtocolCode.GET_WORLD_REFERENCE,
            ]:
                if res:
                    r = []
                    for idx in range(3):
                        r.append(self._int2coord(res[idx]))
                    for idx in range(3, 6):
                        r.append(self._int2angle(res[idx]))
                    return r
                else:
                    return res
            elif genre in [ProtocolCode.GET_SERVO_VOLTAGES]:
                return [self._int2coord(angle) for angle in res]
            elif genre in [ProtocolCode.GET_BASIC_VERSION, ProtocolCode.SOFTWARE_VERSION, ProtocolCode.GET_ATOM_VERSION]:
                return self._int2coord(self._process_single(res))
            elif genre in [
                ProtocolCode.GET_JOINT_MAX_ANGLE,
                ProtocolCode.GET_JOINT_MIN_ANGLE,
            ]:
                return self._int2coord(res[0])
            elif genre == ProtocolCode.GET_ANGLES_COORDS:
                r = []
                for index in range(len(res)):
                    if index < 7:
                        r.append(self._int2angle(res[index]))
                    elif index < 10:
                        r.append(self._int2coord(res[index]))
                    else:
                        r.append(self._int2angle(res[index]))
                return r
            elif genre == ProtocolCode.GO_ZERO:
                r = []
                if res:
                    if 1 not in res[1:]:
                        return res[0]
                    else:
                        for i in range(1, len(res)):
                            if res[i] == 1:
                                r.append(i)
                return r
            elif genre == ProtocolCode.COBOTX_GET_ANGLE:
                    return self._int2angle(res[0])
            else:
                return res
        return None

    def set_solution_angles(self, angle, speed):
        """Set zero space deflection angle value

        Args:
            angle: Angle of joint 1. The angle range is -90 ~ 90
            speed: 1 - 100.
        """
        self.calibration_parameters(
            class_name=self.__class__.__name__, speed=speed, solution_angle=angle
        )
        return self._mesg(
            ProtocolCode.COBOTX_SET_SOLUTION_ANGLES, [self._angle2int(angle)], speed
        )

    def get_solution_angles(self):
        """Get zero space deflection angle value"""
        return self._mesg(ProtocolCode.COBOTX_GET_SOLUTION_ANGLES, has_reply=True)

    def write_move_c(self, transpoint, endpoint, speed):
        """_summary_

        Args:
            transpoint (list): Arc passing point coordinates
            endpoint (list): Arc end point coordinates
            speed (int): 1 ~ 100
        """
        start = []
        end = []
        for index in range(6):
            if index < 3:
                start.append(self._coord2int(transpoint[index]))
                end.append(self._coord2int(endpoint[index]))
            else:
                start.append(self._angle2int(transpoint[index]))
                end.append(self._angle2int(endpoint[index]))
        return self._mesg(ProtocolCode.WRITE_MOVE_C, start, end, speed)

    def focus_all_servos(self):
        """Lock all joints"""
        return self._mesg(ProtocolCode.FOCUS_ALL_SERVOS)

    def go_zero(self):
        """Control the machine to return to the zero position.
        
        Return:
            1 : All motors return to zero position.
            list : Motor with corresponding ID failed to return to zero.
        """
        return self._mesg(ProtocolCode.GO_ZERO, has_reply=True)

    def get_angle(self, joint_id):
        """Get single joint angle

        Args:
            joint_id (int): 1 ~ 7 or 11 ~ 13.
        """
        self.calibration_parameters(class_name=self.__class__.__name__, id=joint_id)
        return self._mesg(ProtocolCode.COBOTX_GET_ANGLE, joint_id, has_reply=True)

    def is_gone_zero(self):
        """Check if it has returned to zero

        Return:
            1 : Return to zero successfully.
            0 : Returning to zero.
        """
        return self._mesg(ProtocolCode.COBOTX_IS_GO_ZERO, has_reply=True)

    def set_encoder(self, joint_id, encoder):
        """Set a single joint rotation to the specified potential value.

        Args:
            joint_id (int): Joint id 1 - 7.
            encoder: The value of the set encoder.
        """
        # TODO CobotX此接口待定
        # self.calibration_parameters(
        #     class_name=self.__class__.__name__, id=joint_id, encoder=encoder
        # )
        return self._mesg(ProtocolCode.SET_ENCODER, joint_id, [encoder])
    
    def servo_restore(self, joint_id):
        """Abnormal recovery of joints

        Args:
            joint_id (int): Joint ID.
                arm : 1 ~ 7 
                waist : 13
                All joints: 254
        """
        self.calibration_parameters(
            class_name=self.__class__.__name__, servo_restore=joint_id
        )
        self._mesg(ProtocolCode.SERVO_RESTORE, joint_id)



# coding=utf-8

from __future__ import division
import time
import math
import socket
import logging

from pymycobot.log import setup_logging
from pymycobot.Interface import MyBuddyCommandGenerator
from pymycobot.common import ProtocolCode, write, read
from pymycobot.error import calibration_parameters


class MyBuddySocket(MyBuddyCommandGenerator):
    """MyCobot Python API Serial communication class.

    Supported methods:

        # Overall status
            Look at parent class: `CommandGenerator`.

        # MDI mode and operation
            get_radians()
            send_radians()
            sync_send_angles() *
            sync_send_coords() *
            Other look at parent class: `CommandGenerator`.

        # JOG mode and operation
            Look at parent class: `CommandGenerator`.

        # Running status and Settings
            Look at parent class: `CommandGenerator`.

        # Servo control
            Look at parent class: `CommandGenerator`.

        # Atom IO
            Look at parent class: `CommandGenerator`.

        # Basic
            Look at parent class: `CommandGenerator`.

        # Other
            wait() *
    """
    _write = write
    _read = read

    def __init__(self, ip, netport=9000):
        """
        Args:
            ip: Server ip
            netport: Server port
        """
        super(MyBuddySocket, self).__init__()
        self.calibration_parameters = calibration_parameters
        self.SERVER_IP = ip
        self.SERVER_PORT = netport
        self.rasp = False
        self.sock = self.connect_socket()
        
    def connect(self, serialport="/dev/ttyAMA0", baudrate="1000000", timeout='0.1'):
        """Connect the robot arm through the serial port and baud rate
        Args:
            serialport: (str) default /dev/ttyAMA0
            baudrate: default 1000000
            timeout: default 0.1
        
        """
        self.rasp = True
        self._write(serialport, "socket")
        self._write(baudrate, "socket")
        self._write(timeout, "socket")

    def connect_socket(self):
        sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        sock.connect((self.SERVER_IP, self.SERVER_PORT))
        return sock

    def _mesg(self, genre, *args, **kwargs):
        """

        Args:
            genre: command type (Command)
            *args: other data.
                   It is converted to octal by default.
                   If the data needs to be encapsulated into hexadecimal,
                   the array is used to include them. (Data cannot be nested)
            **kwargs: support `has_reply`
                has_reply: Whether there is a return value to accept.
        """
        real_command, has_reply = super(
            MyBuddySocket, self)._mesg(genre, *args, **kwargs)
        data = self._write(self._flatten(real_command), "socket")
        if data:
            res = self._process_received(data, genre, arm=12)
            if genre in [
                ProtocolCode.ROBOT_VERSION,
                ProtocolCode.IS_POWER_ON,
                ProtocolCode.IS_CONTROLLER_CONNECTED,
                ProtocolCode.IS_PAUSED,  # TODO have bug: return b''
                ProtocolCode.IS_IN_POSITION,
                ProtocolCode.IS_MOVING,
                ProtocolCode.IS_SERVO_ENABLE,
                ProtocolCode.IS_ALL_SERVO_ENABLE,
                ProtocolCode.GET_SERVO_DATA,
                ProtocolCode.GET_DIGITAL_INPUT,
                ProtocolCode.GET_GRIPPER_VALUE,
                ProtocolCode.IS_GRIPPER_MOVING,
                ProtocolCode.GET_SPEED,
                ProtocolCode.GET_ENCODER,
                ProtocolCode.GET_BASIC_INPUT,
                ProtocolCode.GET_TOF_DISTANCE,
                ProtocolCode.GET_END_TYPE,
                ProtocolCode.GET_MOVEMENT_TYPE,
                ProtocolCode.GET_REFERENCE_FRAME,
                ProtocolCode.GET_JOINT_MIN_ANGLE,
                ProtocolCode.GET_JOINT_MAX_ANGLE
            ]:
                return self._process_single(res)
            elif genre in [ProtocolCode.GET_ANGLES]:
                return [self._int2angle(angle) for angle in res]
            elif genre in [ProtocolCode.GET_COORDS, ProtocolCode.GET_TOOL_REFERENCE, ProtocolCode.GET_WORLD_REFERENCE]:
                if res:
                    r = []
                    for idx in range(3):
                        r.append(self._int2coord(res[idx]))
                    for idx in range(3, 6):
                        r.append(self._int2angle(res[idx]))
                    return r
                else:
                    return res
            elif genre in [ProtocolCode.GET_SERVO_VOLTAGES]:
                return [self._int2coord(angle) for angle in res]
            else:
                return res
        return None

    def get_radians(self, id):
        """Get the radians of all joints

        Args: 
            id: 1/2 (L/R)
            
        Return:
            list: A list of float radians [radian1, ...]
        """
        angles = self._mesg(ProtocolCode.GET_ANGLES, id, has_reply=True)
        return [round(angle * (math.pi / 180), 3) for angle in angles]

    def send_radians(self, id, radians, speed):
        """Send the radians of all joints to robot arm

        Args:
            id: 1/2 (L/R).
            radians: a list of radian values( List[float]), length 6
            speed: (int )0 ~ 100
        """
        # calibration_parameters(len6=radians, speed=speed)
        degrees = [self._angle2int(radian * (180 / math.pi))
                   for radian in radians]
        return self._mesg(ProtocolCode.SEND_ANGLES, id, degrees, speed)

    def set_gpio_mode(self, mode):
        """Set pin coding method
        Args:
            mode: (str) BCM or BOARD 
        """
        self.calibration_parameters(gpiomode=mode)
        if mode == "BCM":
            return self._mesg(ProtocolCode.SET_GPIO_MODE, 0)
        else:
            return self._mesg(ProtocolCode.SET_GPIO_MODE, 1)

    def set_gpio_out(self, pin_no, mode):
        """Set the pin as input or output
        Args:
            pin_no: (int) pin id
            mode: (str) "in" or "out"
        """
        if mode == "in":
            return self._mesg(ProtocolCode.SET_GPIO_UP, pin_no, 0)
        else:
            return self._mesg(ProtocolCode.SET_GPIO_UP, pin_no, 1)

    def set_gpio_output(self, pin_no, state):
        """Set the pin to high or low level
        Args:
            pin_no: (int) pin id.
            state: (int) 0 or 1
        """
        return self._mesg(ProtocolCode.SET_GPIO_OUTPUT, pin_no, state)

    def get_gpio_in(self, pin_no):
        """Get pin level status.
        Args:
            pin_no: (int) pin id.
        """
        return self._mesg(ProtocolCode.GET_GPIO_IN, pin_no)

    # Other
    def wait(self, t):
        time.sleep(t)
        return self
    
    def close(self):
        self.sock.close()





import sys
import logging
import time

from pymycobot.log import setup_logging
from pymycobot.error import calibration_parameters
from pymycobot.common import ProtocolCode, DataProcessor


class PublicCommandGenerator(DataProcessor):
    # def __init__(self, debug=False):
    #     """
    #     Args:
    #         debug    : whether show debug info
    #     """
    #     self._version = sys.version_info[:2][0]
    #     self.debug = debug
    #     setup_logging(self.debug)
    #     self.log = logging.getLogger(__name__)
    #     self.calibration_parameters = calibration_parameters

    # def _mesg(self, genre, *args, **kwargs):
    #     """
    #     Args:
    #         genre: command type (Command)
    #         *args: other data.
    #                It is converted to octal by default.
    #                If the data needs to be encapsulated into hexadecimal,
    #                the array is used to include them. (Data cannot be nested)
    #         **kwargs: support `has_reply`
    #             has_reply: Whether there is a return value to accept.
    #     """
    #     command_data = self._process_data_command(genre, args)

    #     if genre == 178:
    #         command_data = self._encode_int16(command_data)
            
    #     elif genre in [76, 77]:
    #         command_data = [command_data[0]] + self._encode_int16(command_data[1]*10)
    #     elif genre == 115:
    #         command_data = [command_data[1],command_data[3]]

    #     LEN = len(command_data) + 2
    #     command = [
    #         ProtocolCode.HEADER,
    #         ProtocolCode.HEADER,
    #         LEN,
    #         genre,
    #         command_data,
    #         ProtocolCode.FOOTER,
    #     ]

    #     real_command = self._flatten(command)
    #     has_reply = kwargs.get("has_reply", False)
    #     return real_command, has_reply
    pass
